import regex as regex_module
import logging
from .._guidance import guidance
from ._silent import silent
from .._grammar import select
from ._sequences import zero_or_more
from .._grammar import commit_point
from ._any_char import any_char
from .._grammar import capture
from ._regex import regex as regex_grammar
from .._grammar import token_limit, eos_token, active_role_end, with_temperature
from ._tool import Tool
from ._block import block

logger = logging.getLogger(__name__)


# TODO: make this stateless!
# TODO: uncomment this once we get temperature stateless
@guidance(stateless=lambda *args, **kwargs: kwargs.get("tools", None) is None)  
def gen(
    lm,
    name=None,
    *,
    max_tokens=1e10,
    list_append=False,
    regex=None,
    tools=None,
    hide_tool_call=False,
    stop=None,
    stop_regex=None,
    suffix="",
    n=1,
    temperature=0.0,
    top_p=1.0,
    save_stop_text=False,
):
    """Generate a set of tokens until a given stop criteria has been met.

    This function is a useful utility that can allow you to specify most grammars used by typical
    LM generation programs. It also has the added ability to interleave generation with tool calls.

        >>> lm += gen("my_generation", max_tokens=10)
        >>> print(lm["my_generation"])
        some text from the LLM

    Parameters
    ----------

        name : str or None
            If this is not None then the the results of the generation will be saved as a variable on
            the Model object (so you can access the result as `lm["var_name"]`).

        max_tokens : int
            The maximum number of generation tokens we should use. Note that this limit is not exact when
            regular expression pattern constraints are present, but guidance does attempt to end the generation
            as soon as possible while keeping the regex constraints satisfied.

        list_append : bool
            If this is True then the results saved to `lm[name]` will not be written directly but rather appended
            to a list (if no list with the current name is present one will be created). This is useful for
            building lists inside python loops.

        regex : str or None
            This is a regular expression that will be used to constrain the generation. The model is only allowed
            to generate tokens that match this regular expression. Note that for variable length expressions the
            model is free to continue the expression after a complete match, but generation will terminate as soon
            as the model generates anything that does not match the pattern (this ending behavior may change a bit we
            update guidance to maintain the grammar parsing state between calls).

        stop : str or list or None
            The stop string (or list of strings) we should use for terminating this generation segment.

        stop_regex : str or list or None
            The stop regular expression (or list of regular expressions) we should use for terminating this generation segment.

        save_stop_text : bool or str
            If True then this saves the captured stop text or regex into a variable of the name `str(name) + "_stop_text"`. If
            a string is given then the captured stop text is saved under that name.

        temperature : float
            The temperature to use during this generation call. Note that when parsing ambiguous grammars that include
            multiple conflicting temperatures (for example from multiple possible `gen` calls inside a `select`) the highest
            temperature of all options is used by the model (since we only want to run the model once, not once for every
            possible parse path).

        top_p : float
            TODO! Will control the models top_p generation parameter, but has been yet been implemented beyond top_p=1.0.

        n : int
            TODO! Will control the number of parallel generation calls made during gen.

        tools : Tool or list or None
            A list of guidance.Tool or python functions (which will be converted to guidance.Tool)

        hide_tool_call : bool
            Controls if we should hide the text generated by the model to trigger a tool call. You may want to hide the tool
            call from the model's context if you plan to change it's format after the call is made.
    """
    # TODO: expand the tools doc string
    assert (
        n == 1
    ), "We still need to add support for n>1! Consider putting your gen call in a loop for now."
    assert top_p == 1, "We still need to add support for top_p != 1!"

    logger.debug(f'start gen(name="{name}")')

    # set stream if we are interactive
    # if stream_tokens is None and not lm.is_silent() and n == 1:
    #     stream_tokens = True

    # use the suffix as the stop string if not otherwise specified
    # TODO: still need to make suffix work with grammars
    # eos_token = lm.eos_token.decode('utf8')
    if stop is None and stop_regex is None and suffix != "":
        stop = suffix
    # if stop is None and stop_regex is None and getattr(lm, "suffix", False):
    #     if lm.suffix.startswith("\n"):
    #         stop = "\n"
    #     elif lm.suffix.startswith('"') and str(lm).endswith('"'):
    #         stop = '"'
    #     elif lm.suffix.startswith("'") and str(lm).endswith("'"):
    #         stop = "'"

    # fall back to stopping at the EOS token
    if stop is not False:
        if stop is None:
            stop = []
        if isinstance(stop, str):
            stop = [stop]
        if regex is None:
            stop = stop + [select([eos_token(), active_role_end()])]

        if stop_regex is None:
            stop_regex = []
        if isinstance(stop_regex, str):
            stop_regex = [stop_regex]
        stop_regex = [regex_grammar(x) for x in stop_regex]

    # This needs to be here for streaming
    # if name is not None and not list_append:
    #     lm[name] = ""

    # define the generation pattern
    if regex is not None:
        pattern = regex_grammar(regex)
    else:
        pattern = zero_or_more(any_char())

    tagged_name = "__LIST_APPEND:" + name if list_append and name is not None else name

    # define any capture group for non-tool calls
    if name is not None and tools is None:
        pattern = capture(pattern, name=tagged_name)

    # limit the number of tokens
    pattern = token_limit(pattern, max_tokens)

    # define the stop pattern
    if stop is False or len(stop + stop_regex) == 0:
        stop_pattern = ""
    else:
        stop_pattern = select(stop + stop_regex)
        if save_stop_text is True:
            save_stop_text = str(name) + "_stop_text"
        if isinstance(save_stop_text, str):
            stop_pattern = capture(stop_pattern, name=save_stop_text)
        stop_pattern = commit_point(stop_pattern, hidden=True)

    # single generation
    start_pos = len(str(lm))
    if tools is not None:
        with block(tagged_name):
            tools = [Tool(callable=x) if not isinstance(x, Tool) else x for x in tools]
            init_token_count = lm.token_count
            gen_grammar = pattern + select(
                [stop_pattern]
                + [
                    capture(
                        commit_point(x.call_grammar, hidden=hide_tool_call),
                        name=f"tool{i}",
                    )
                    for i, x in enumerate(tools)
                ]
            )
            while lm.token_count <= max_tokens + init_token_count:
                lm = lm._run_stateless(
                    gen_grammar, temperature=temperature
                )  # TODO: we should not be using this internal method
                tool_called = False
                for i in range(len(tools)):
                    tool_i = f"tool{i}"
                    if tool_i in lm:
                        tool_called = True
                        lm += tools[i].tool_call()
                        lm = lm.remove(tool_i)
                if not tool_called:
                    lm += suffix
                    break
    elif n == 1:
        lm += with_temperature(pattern + stop_pattern + suffix, temperature)

    logger.debug(f"finish gen")
    return lm


def click_loop_start(id, total_count, echo, color):
    click_script = (
        """
function cycle_IDVAL(button_el) {
var i = 0;
while (i < 50) {
var el = document.getElementById("IDVAL_" + i);
if (el.style.display == "inline") {
    el.style.display = "none";
    var next_el = document.getElementById("IDVAL_" + (i+1));
    if (!next_el) {
        next_el = document.getElementById("IDVAL_0");
    }
    if (next_el) {
        next_el.style.display = "inline";
    }
    break;
}
i += 1;
}
button_el.innerHTML = (((i+1) % TOTALCOUNT) + 1)  + "/" + TOTALCOUNT;
}
cycle_IDVAL(this);""".replace(
            "IDVAL", id
        )
        .replace("TOTALCOUNT", str(total_count))
        .replace("\n", "")
    )
    out = f"""<div style='background: rgba(255, 255, 255, 0.0); border-radius: 4px 0px 0px 4px; border: 1px solid {color}; border-right: 0px; padding-left: 3px; padding-right: 3px; user-select: none; color: {color}; display: inline; font-weight: normal; cursor: pointer' onClick='{click_script}'>1/{total_count}</div>"""
    out += f"<div style='display: inline;' id='{id}_0'>"
    return "<||_html:" + out + "_||>"


def click_loop_mid(id, index, echo):
    alpha = 1.0 if not echo else 0.5
    out = f"</div><div style='display: none; opacity: {alpha}' id='{id}_{index}'>"
    return "<||_html:" + out + "_||>"


@guidance
def gen_line(lm, *args, **kwargs):
    return lm.gen(*args, suffix="\n", **kwargs)


@guidance
def gen_quote(lm, name=None, quote='"', *args, **kwargs):
    return lm(quote).gen(*args, name=name, suffix=quote, **kwargs)


@guidance
def will_gen(lm, stop=None, stop_regex=None, ignore_spaces=False, max_tokens=30):
    # this is obviously not the right implementation, just here so we can explore
    if stop and not isinstance(stop, list):
        stop = [stop]
    if stop_regex and not isinstance(stop_regex, list):
        stop_regex = [stop_regex]
    assert (stop is not None) or (stop_regex is not None)
    if not stop:
        stop = []
    if not stop_regex:
        stop_regex = []
    regexes = [regex_module.escape(x) for x in stop + stop_regex]
    optional_space = "\\s*" if ignore_spaces else ""
    pattern = regex_module.compile(f'{optional_space}({"|".join(regexes)})')
    lm2 = lm
    with silent():
        for _ in range(max_tokens):
            lm2 += gen("temp_variable", list_append=True, max_tokens=1)
            if not lm2["temp_variable"] or not pattern.match(
                "".join(lm2["temp_variable"]), partial=True
            ):
                return False
            if pattern.match("".join(lm2["temp_variable"]), partial=False):
                return True
    return False


@guidance
def call_tool(lm, tool):
    return lm + tool.call_grammar + tool.tool_call()



import functools
import inspect

from . import models
from ._grammar import Placeholder, RawFunction, Terminal, replace_grammar_node, string
from ._utils import strip_multiline_string_indents


def guidance(f=None, *, stateless=False, cache=None, dedent=True, model=models.Model):
    return _decorator(f, stateless=stateless, cache=cache, dedent=dedent, model=model)


_null_grammar = string("")


def _decorator(f, *, stateless, cache, dedent, model):

    # if we are not yet being used as a decorator, then save the args
    if f is None:
        return functools.partial(
            _decorator, stateless=stateless, cache=cache, dedent=dedent, model=model
        )

    # if we are being used as a decorator then return the decorated function
    else:

        # this strips out indentation in multiline strings that aligns with the current python indentation
        if dedent is True or dedent == "python":
            f = strip_multiline_string_indents(f)

        # we cache if requested
        if cache:
            f = functools.cache(f)

        @functools.wraps(f)
        def wrapped(*args, **kwargs):

            # make a stateless grammar if we can
            if stateless is True or (
                callable(stateless) and stateless(*args, **kwargs)
            ):

                # if we have a placeholder set then we must be in a recursive definition and so we return the placeholder
                placeholder = getattr(f, "_self_call_placeholder_", None)
                if placeholder is not None:
                    return placeholder

                # otherwise we call the function to generate the grammar
                else:

                    # set a placeholder for recursive calls (only if we don't have arguments that might make caching a bad idea)
                    no_args = len(args) + len(kwargs) == 0
                    if no_args:
                        f._self_call_placeholder_ = Placeholder()

                    # call the function to get the grammar node
                    node = f(_null_grammar, *args, **kwargs)
                    if not isinstance(node, (Terminal, str)):
                        node.name = f.__name__

                    # replace all the placeholders with our generated node
                    if no_args:
                        replace_grammar_node(node, f._self_call_placeholder_, node)
                        del f._self_call_placeholder_

                    return node

            # otherwise must be stateful (which means we can't be inside a select() call)
            else:
                return RawFunction(f, args, kwargs)

        # Remove the first argument from the wrapped function
        signature = inspect.signature(f)
        params = list(signature.parameters.values())
        params.pop(0)
        wrapped.__signature__ = signature.replace(parameters=params)

        # attach this as a method of the model class (if given)
        # if model is not None:
        #     setattr(model, f.__name__, f)

        return wrapped



import http
import pathlib
import re
import typing
import urllib

from .._guidance import guidance


@guidance
def image(lm, src: typing.Union[str, pathlib.Path, bytes], allow_local: bool = True):

    # load the image bytes
    # ...from a url
    if isinstance(src, str) and re.match(r"[^:/]+://", src):
        with urllib.request.urlopen(src) as response:
            response = typing.cast(http.client.HTTPResponse, response)
            bytes_data = response.read()

    # ...from a local path
    elif allow_local and (isinstance(src, str) or isinstance(src, pathlib.Path)):
        with open(src, "rb") as f:
            bytes_data = f.read()

    # ...from image file bytes
    elif isinstance(src, bytes):
        bytes_data = src

    else:
        raise Exception(f"Unable to load image bytes from {src}!")

    bytes_id = str(id(bytes_data))

    # set the image bytes
    lm = lm.set(bytes_id, bytes_data)
    lm += f"<|_image:{bytes_id}|>"
    return lm



from typing import Any, Dict, Sequence, Union

import numpy as np

from ..chat import load_template_class, ChatTemplate


class Tokenizer:
    """This is the standardized tokenizer interface used by guidance models.

    This class should be subclassed by specific implementations and then used as the
    tokenizer in the corresponding Engine subclass.
    """

    def __init__(
        self,
        tokens: Union[Sequence[bytes], np.ndarray],
        chat_template: Union[str, ChatTemplate, None],
        bos_token_id: Union[int, None] = None,
        eos_token_id: Union[int, None] = None,
    ):

        # a numpy array of token byte strings indexed by their token id
        if isinstance(tokens, list):
            # note that we need np.bytes_ so zero bytes are not treated as null terminations
            self._tokens = np.array(tokens, dtype="object")

        # a numpy array of token byte strings indexed by their token id
        elif isinstance(tokens, np.ndarray):
            self._tokens = tokens

        else:
            raise ValueError("Unknown tokenizer was passed!")

        assert isinstance(
            self.tokens[0], bytes
        ), "The tokens need to be provided as bytes!"

        # This method supports None, a huggingface style jinja2_template_str, or a ChatTemplate subclass
        # Defaults to ChatML if nothing is found
        self._chat_template = load_template_class(chat_template)

        self._bos_token_id = bos_token_id
        self._bos_token = (
            None if self.bos_token_id is None else self.tokens[self.bos_token_id]
        )
        self._eos_token_id = eos_token_id if eos_token_id is not None else bos_token_id
        self._eos_token = (
            None if self.eos_token_id is None else self.tokens[self.eos_token_id]
        )

        # track which tokens are duplicates
        self._duplicate_tokens = []
        found: Dict[bytes, int] = {}
        for i, t in enumerate(self.tokens):
            if t in found:
                self._duplicate_tokens.append((i, found[t]))
            else:
                found[t] = i

    @property
    def tokens(self) -> np.ndarray:
        return self._tokens

    @property
    def bos_token_id(self) -> Union[int, None]:
        return self._bos_token_id

    @property
    def eos_token_id(self) -> Union[int, None]:
        return self._eos_token_id

    @property
    def bos_token(self) -> Union[bytes, None]:
        return self._bos_token

    @property
    def eos_token(self) -> Union[bytes, None]:
        return self._eos_token

    @property
    def chat_template(self) -> Union[Any, None]:
        return self._chat_template

    def __call__(self, byte_string: bytes):
        return self.encode(byte_string)

    def encode(self, byte_string: bytes) -> Sequence[int]:
        """Returns a list of tokens that represent the given byte string."""
        raise NotImplementedError(
            "You need to use a Tokenize subclass that overrides the encode method"
        )

    def decode(self, tokens: Sequence[int]) -> bytes:
        """Returns the bytes represented by the given list of tokens."""
        return b"".join([self.tokens[t] for t in tokens])

    def recode(self, tokens: Sequence[int]) -> Sequence[int]:
        """Redoes a tokenisation.

        Encoding a string into tokens does not distribute over concatenation.
        That is, in general, `encode(A)+encode(B) != encode(A+B)` (although it
        it may in some cases).
        An LLM will generate token-by-token, but it is possible (even likely) that
        when the generation is considered as a whole, a better tokenisation may
        be possible.
        This method takes in a sequence of tokens, and returns an 'improved' sequence.
        """

        # This is the notional behaviour
        # It may need to be overridden in particular cases because
        # we are dealing with LLM ecosystems in the real world
        return self.encode(self.decode(tokens))

    def clean_duplicate_tokens(self, probs):
        """This moves all the probability mass from duplicate positons on to their primary index."""
        for i, j in self._duplicate_tokens:
            probs[j] += probs[i]
            probs[i] = 0



from typing import Union

from .._grammar import Join, select
from .._guidance import guidance
from ._optional import optional


@guidance(stateless=True)
def exactly_n_repeats(model, value, n_repeats: int):
    assert n_repeats >= 0, f"Must have n_repeats>=0 ({n_repeats})"
    return model + Join([value] * n_repeats)


@guidance(stateless=True)
def at_most_n_repeats(model, value, n_repeats: int):
    assert n_repeats >= 0

    return model + exactly_n_repeats(optional(value), n_repeats)


@guidance(stateless=True)
def sequence(model, value, min_length: int = 0, max_length: Union[int, None] = None):
    assert min_length >= 0
    assert max_length is None or max_length >= min_length

    if max_length is not None:
        model += exactly_n_repeats(value=value, n_repeats=min_length)
        model += at_most_n_repeats(value=value, n_repeats=(max_length - min_length))
    else:
        model += exactly_n_repeats(value=value, n_repeats=min_length)
        model += select([optional(value)], recurse=True)
    return model


@guidance(stateless=True)
def one_or_more(model, value):
    return model + sequence(value, min_length=1)


@guidance(stateless=True)
def zero_or_more(model, value):
    return model + select(["", value], recurse=True)



from guidance import models


class ContextBlock:
    def __init__(self, opener, closer, name=None):
        self.opener = opener
        self.closer = closer
        self.name = name

    def __enter__(self):
        models.Model.open_blocks[self] = None

    def __exit__(self, exc_type, exc_value, traceback):
        del models.Model.open_blocks[self]


def block(name=None, opener="", closer=""):
    return ContextBlock(opener, closer, name=name)



import inspect
from typing import Any, Dict, Type, TYPE_CHECKING, Union

import pydantic


class GenerateJsonSchemaSafe(pydantic.json_schema.GenerateJsonSchema):
    """
    Subclass pydantic's GenerateJsonSchema to catch pydantic schemas that will not
    translate properly to json schemas used for generation.

    In particular, JSON schemas do not offer a way to specify "key type",
    so we need to raise an exception if users attempt to specify non-string
    keys through pydantic. Otherwise, they may get unexpected output from
    model generation.
    """

    def generate_inner(self, schema):
        if schema["type"] == "dict":
            key_type = schema["keys_schema"]["type"]
            if key_type != "str":
                raise TypeError(
                    f"JSON does not support non-string keys, got type {key_type}"
                )
        return super().generate_inner(schema)


def pydantic_to_json_schema(
    schema: Union[Type["pydantic.BaseModel"], "pydantic.TypeAdapter"]
) -> Dict[str, Any]:
    if inspect.isclass(schema) and issubclass(schema, pydantic.BaseModel):
        return schema.model_json_schema(schema_generator=GenerateJsonSchemaSafe)
    if isinstance(schema, pydantic.TypeAdapter):
        return schema.json_schema(schema_generator=GenerateJsonSchemaSafe)
    raise TypeError(f"Cannot generate json schema from type {type(schema)}")



from .._guidance import guidance
from .._grammar import select, string


@guidance(stateless=True, dedent=False)
def prefix_tree(lm, strings, partial_matches=False):

    if len(strings) == 0:
        return lm

    # group the strings by their starting character
    char_groups = {}
    for s in strings:
        if len(s) > 0:
            if s[0] not in char_groups:
                char_groups[s[0]] = []
            char_groups[s[0]].append(s[1:])

    # enable any empty followup if partial matches are allowed
    if partial_matches:
        char_groups[""] = []

    # recursively build the tree
    suboptions = [
        string(k) + prefix_tree(v, partial_matches=partial_matches)
        for k, v in char_groups.items()
    ]

    return lm + select(
        suboptions, skip_checks=True
    )  # we skip normal type checks for speed



from .._guidance import guidance
from ._block import block


@guidance
def set_opener(lm, name, value):
    if name in lm:
        lm = lm.set("__save" + name, lm[name])
    return lm.set(name, value)


@guidance
def set_closer(lm, name):
    if "__save" + name in lm:
        return lm.set(name, lm["__save" + name]).remove("__save" + name)
    else:
        return lm.remove(name)


def set_var(name, value=True):
    return block(
        opener=set_opener(name, value),
        closer=set_closer(name),
    )



import os
from pathlib import Path
import multiprocessing
from itertools import takewhile
import operator
import threading
import numpy as np
import queue
import time
import tiktoken

from ._vertexai import VertexAICompletion, VertexAIInstruct, VertexAIChat

try:
    from vertexai.language_models import (
        TextGenerationModel,
        ChatModel,
        InputOutputTextPair,
    )

    is_vertexai = True
except ModuleNotFoundError:
    is_vertexai = False


class PaLM2Completion(VertexAICompletion):
    def __init__(
        self, model, tokenizer=None, echo=True, max_streaming_tokens=None, **kwargs
    ):

        if isinstance(model, str):
            model = TextGenerationModel.from_pretrained(self.model_name)

        # PaLM2 does not have a public tokenizer, so we pretend it tokenizes like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")

        # the superclass does all the work
        super().__init__(
            model,
            tokenizer=tokenizer,
            echo=echo,
            max_streaming_tokens=max_streaming_tokens,
            **kwargs,
        )


class PaLM2Instruct(VertexAIInstruct):
    def __init__(
        self, model, tokenizer=None, echo=True, max_streaming_tokens=None, **kwargs
    ):
        if isinstance(model, str):
            model = TextGenerationModel.from_pretrained(model)

        # PaLM2 does not have a public tokenizer, so we pretend it tokenizes like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")

        # the superclass does all the work
        super().__init__(
            model,
            tokenizer=tokenizer,
            echo=echo,
            max_streaming_tokens=max_streaming_tokens,
            **kwargs,
        )


class PaLM2Chat(VertexAIChat):
    def __init__(
        self, model, tokenizer=None, echo=True, max_streaming_tokens=None, **kwargs
    ):
        if isinstance(model, str):
            model = ChatModel.from_pretrained(model)

        # PaLM2 does not have a public tokenizer, so we pretend it tokenizes like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")

        # the superclass does all the work
        super().__init__(
            model,
            tokenizer=tokenizer,
            echo=echo,
            max_streaming_tokens=max_streaming_tokens,
            **kwargs,
        )



import sys

if sys.version_info >= (3, 11):
    import re._constants as constants  # type: ignore[import-not-found]
    import re._parser as parser  # type: ignore[import-not-found]
else:
    import sre_parse as parser
    import sre_constants as constants

from re import RegexFlag
from typing import Any, List, Tuple, Union

from typing_extensions import TypeAlias

from .._grammar import Byte, ByteRange, Join, Select, byte_range, select, capture
from .._guidance import guidance
from ._any_char_but import any_char_but
from ._sequences import sequence

# Type aliases
Node: TypeAlias = Tuple[constants._NamedIntConstant, Any]


class UnsupportedRegexError(Exception):
    pass


class RegexPatternConverter:

    @classmethod
    def parse(cls, pattern: str):
        return cls.convert(parser.parse(pattern))

    @classmethod
    def convert(cls, tree: Union[parser.SubPattern, Node], flags: int = 0):
        if flags != 0:
            # Equivalent to re.NOFLAG
            raise UnsupportedRegexError(
                f"Flags other than re.NOFLAG not supported; got {RegexFlag(flags)}"
            )
        if isinstance(tree, parser.SubPattern):
            if len(tree.data) == 1:
                return cls.convert(tree.data[0])
            return Join([cls.convert(node) for node in tree.data])

        opcode, args = tree
        opcode_name = opcode.name
        try:
            method = getattr(cls, opcode_name)
        except AttributeError as e:
            raise UnsupportedRegexError(
                f"Unsupported regex feature with opcode {opcode_name}"
            ) from e
        return method(args)

    @classmethod
    def SUBPATTERN(cls, args: Tuple[int, int, int, parser.SubPattern]):
        # capture group
        group, add_flags, del_flags, arg = args
        flags = add_flags & ~del_flags
        return cls.convert(arg, flags)

    @classmethod
    def LITERAL(cls, args: int):
        # byte
        return Byte(bytes([args]))

    @classmethod
    def NOT_LITERAL(cls, args: int):
        return any_char_but(chr(args))

    @classmethod
    def RANGE(cls, args: Tuple[int, int]):
        # byte_range
        low, high = args
        return byte_range(bytes([low]), bytes([high]))

    @classmethod
    def ANY(cls, _: None):
        return any_char_but("\n")

    @classmethod
    def IN(cls, args: List[Node]):
        if args[0][0] == constants.NEGATE:
            transformed_args = [cls.convert(arg) for arg in args[1:]]
            negated_bytes = cls._get_negated_bytes(transformed_args)
            return any_char_but(negated_bytes)
        transformed_args = [cls.convert(arg) for arg in args]
        return select(transformed_args)

    @classmethod
    def _get_negated_bytes(cls, grammars: List[Union[Byte, ByteRange, Select]]):
        negated_bytes = set()
        for value in grammars:
            if isinstance(value, Byte):
                negated_bytes.add(value.byte)
            elif isinstance(value, ByteRange):
                low, high = value.byte_range
                negated_bytes.update([bytes([i]) for i in range(low, high + 1)])
            elif isinstance(value, Select):
                negated_bytes.update(cls._get_negated_bytes(value._values))
            else:
                raise TypeError(f"Can't negate {type(value)} object")
        return negated_bytes

    @classmethod
    def BRANCH(cls, args: Tuple[Any, List[parser.SubPattern]]):
        unknown, arg = args
        if unknown is not None:
            # Unsure of the semantics of this value, but it seems to be
            # None in all cases tested so far
            raise UnsupportedRegexError(f"Unkwnown argument in BRANCH: {unknown}")
        transformed_args = [cls.convert(a) for a in arg]
        return select(transformed_args)

    @classmethod
    def MAX_REPEAT(
        cls,
        args: Tuple[int, Union[int, constants._NamedIntConstant], parser.SubPattern],
    ):
        low, high, arg = args
        transformed_arg = cls.convert(arg)
        if isinstance(high, constants._NamedIntConstant):
            if high != constants.MAXREPEAT:
                raise UnsupportedRegexError(f"Unsupported high value in range: {high}")
            return sequence(transformed_arg, min_length=low)
        return sequence(transformed_arg, min_length=low, max_length=high)

    @classmethod
    def CATEGORY(cls, args: constants._NamedIntConstant):
        # \d
        if args.name == "CATEGORY_DIGIT":
            return cls.parse(r"[0-9]")
        # \D
        if args.name == "CATEGORY_NOT_DIGIT":
            return cls.parse(r"[^0-9]")
        # \w
        if args.name == "CATEGORY_WORD":
            return cls.parse(r"[0-9A-Za-z_]")
        # \W
        if args.name == "CATEGORY_NOT_WORD":
            return cls.parse(r"[^0-9A-Za-z_]")
        # \s
        if args.name == "CATEGORY_SPACE":
            return cls.parse(r"[ \t\n\r\f\v]")
        # \S
        if args.name == "CATEGORY_NOT_SPACE":
            return cls.parse(r"[^ \t\n\r\f\v]")
        raise UnsupportedRegexError(f"Unsupported category: {args.name}")


@guidance(stateless=True)
def regex(lm, pattern, *, name=None):
    return lm + capture(RegexPatternConverter.parse(pattern), name=name)



from .._guidance import guidance
from .._grammar import capture as grammar_capture, GrammarFunction


@guidance(stateless=lambda *args, **kwargs: isinstance(args[0], GrammarFunction))
def capture(lm, value, name):
    if isinstance(value, GrammarFunction):
        return lm + grammar_capture(value, name)
    else:
        start_len = len(lm)
        lm += value
        return lm.set(name, str(lm)[start_len:])
